import React, { ReactElement } from 'react';
import { ViewProps } from 'react-native';
import { RNINavigatorViewProps } from '../native_components/RNINavigatorView';
import { NavigatorConstantsObject } from '../native_modules/RNINavigatorViewModule';
import type { NavCommandPopOptions, NavCommandPushOptions, RouteTransitionConfig } from '../types/NavigationCommands';
import type { NavRouteItem, NavRouteStackItem, NavRouteStackPartialItem } from '../types/NavRouteItem';
import type { RenderNavItem } from '../types/NavTypes';
import type { NavRouteConfigItem } from '../types/NavRouteConfigItem';
import type { OnUIConstantsDidChangeEventObject } from '../types/RNINavigatorViewEvents';
export declare type SetRoutesTransformCallback = (currentRoutes: readonly NavRouteStackPartialItem[]) => Array<NavRouteStackPartialItem>;
export declare type NavRoutesConfigMap = {
    [k: string]: NavRouteConfigItem;
};
/** `NavigatorView` comp. props */
declare type NavigatorViewBaseProps = Partial<Pick<RNINavigatorViewProps, 'isInteractivePopGestureEnabled' | 'shouldSwizzleRootViewController' | 'navBarPrefersLargeTitles' | 'navBarAppearance' | 'isNavBarTranslucent' | 'disableTransparentNavBarScrollEdgeAppearance' | 'onCustomCommandFromNative' | 'onNavRouteWillPop' | 'onNavRouteDidPop' | 'onNavRouteWillShow' | 'onNavRouteDidShow' | 'onUIConstantsDidChange'>> & {
    routes: NavRoutesConfigMap;
    initialRoutes: Array<NavRouteItem>;
    sharedRouteProps?: object;
    renderNavBarLeftItem?: RenderNavItem;
    renderNavBarRightItem?: RenderNavItem;
    renderNavBarTitleItem?: RenderNavItem;
    renderNavBarBackground?: () => ReactElement;
};
export declare type NavigatorViewProps = ViewProps & NavigatorViewBaseProps;
/** `NavigatorView` comp. state */
declare type NavigatorViewState = Partial<Pick<OnUIConstantsDidChangeEventObject['nativeEvent'], 'statusBarHeight' | 'safeAreaInsets' | 'navigatorSize'>> & {
    activeRoutes: Array<NavRouteStackItem>;
    transitionConfigPushOverride?: RouteTransitionConfig;
    transitionConfigPopOverride?: RouteTransitionConfig;
};
export declare class NavigatorView extends React.PureComponent<NavigatorViewProps, NavigatorViewState> {
    /** A ref to the `RNINavigatorView` native component. */
    nativeRef: React.Component;
    /** Unique identifier for this navigator */
    private navigatorID;
    private navStatus;
    private emitter;
    /** Used for `removeRouteBatchedFromState`*/
    private routesToRemove;
    /** Queue for nav. commands: only 1 command is called at a time */
    private queue;
    private routeRefMap;
    constructor(props: NavigatorViewProps);
    componentWillUnmount(): void;
    private verifyProps;
    private isValidRouteKey;
    private getInitialRoutes;
    /** Creates a dict of the current native routes based on `state.activeRoutes` */
    private getNativeRoutes;
    /** returns the current js/react routes sorted by `routeID` */
    private getRoutesToRender;
    /** get route config with the matching `routeKey` */
    private getRouteConfig;
    private getLastRouteTransitionDuration;
    private setNavStatus;
    /** used to set/reset the transition override for the current route  */
    private configureTransitionOverride;
    /**
     * Returns a promise that will resolve once all the routes in `routeItems` are added/initialized
     * from the native-side.
     * * If the route item in `routeItems` has a native route, then wait for them to be created
     *   and initialized with data (e.g. wait for the `onNavRouteViewAdded` event).
     * * For each react/js route in `routeItems`, wait for it to be added as a subview in the
     *   native side (e.g. wait for the `onNavRouteViewAdded` event).
     *
     * note: This promise will reject if the events fail to fire within `TIMEOUT_MOUNT` ms. */
    private waitForRoutesToBeAdded;
    /**
     * Returns a promise that will resolve once all the routes in `routeItems` are removed
     * from from the native-side (i.e. wait for `onNavRouteDidPop`).
     */
    private waitForRoutesToBeRemoved;
    private syncRoutesFromNative;
    /** Remove route from `state.activeRoutes` in batches a few ms apart. */
    private removeRouteBatchedFromState;
    private _handleGetRefToNavigator;
    getActiveRoutes: () => NavRouteStackItem[];
    push: (routeItem: Readonly<NavRouteItem>, options?: NavCommandPushOptions | undefined) => Promise<void>;
    pop: (options?: Readonly<NavCommandPopOptions> | undefined) => Promise<void>;
    popToRoot: (options?: Readonly<NavCommandPopOptions> | undefined) => Promise<void>;
    removeRoute: (routeIndex: number, animated?: boolean) => Promise<void>;
    removeRoutes: (routeIndices: readonly number[], animated?: boolean) => Promise<void>;
    replaceRoute: (prevRouteIndex: number, routeItem: Readonly<NavRouteItem>, animated?: boolean) => Promise<void>;
    insertRoute: (routeItem: Readonly<NavRouteItem>, atIndex: number, animated?: boolean) => Promise<void>;
    setRoutes: (transform: SetRoutesTransformCallback, animated?: boolean) => Promise<void>;
    setNavigationBarHidden: (isHidden: boolean, animated: boolean) => Promise<void>;
    replacePreviousRoute: (routeItem: Readonly<NavRouteItem>, animated?: boolean) => Promise<void>;
    replaceCurrentRoute: (routeItem: Readonly<NavRouteItem>, animated?: boolean) => Promise<void>;
    removePreviousRoute: (animated?: boolean) => Promise<void>;
    removeAllPrevRoutes: (animated?: boolean) => Promise<void>;
    sendCustomCommandToNative: (commandKey: string, commandData?: Readonly<object> | null) => Promise<object | null>;
    getNavigatorConstants: () => Promise<NavigatorConstantsObject>;
    dismissModal: (animated?: boolean) => Promise<void>;
    private _handleRouteViewWrapperRef;
    private _handleOnNavRouteViewAdded;
    private _handleOnSetNativeRoutes;
    private _handleOnNativeCommandRequest;
    private _handleOnCustomCommandFromNative;
    private _handleOnNavRouteWillPop;
    private _handleOnNavRouteDidPop;
    private _handleOnUIConstantsDidChange;
    private _handleOnNavRouteDidShow;
    private _renderRoutes;
    render(): JSX.Element;
}
export {};
